/**
 * @file scripting.cpp
 * Implementation of Lua class.
 *
 * Copyright 2004 by Gamebuilders.
 * Distributed WITHOUT WARRANTY under the GPL. See COPYING for details.
 */
#include "scripting.h"
#include "main.h"

#include "common.hpp"

extern "C" {
#include <lua.h>
#include <lualib.h>
#include <lauxlib.h>
}

#include <sstream>

#include <windows.h>

// C++ defined Lua functions

static int l_message_box( lua_State* state )
{
    if ( !lua_isprintable( state, 1 ) ) {
        lua_pushstring( state, "incorrect argument list: message_box" );
        lua_error( state );
    }

    std::string text;
    if ( lua_isboolean( state, 1 ) ) {
       text = ( lua_toboolean( state, 1 ) ? "true" : "false" );
    }
    else {
       text = lua_tostring( state, 1 );
    } 

    std::string title = "LSLua";
    if ( lua_isprintable( state, 2 ) ) {
        if ( lua_isboolean( state, 2 ) ) {
            title = ( lua_toboolean( state, 2 ) ? "true" : "false" );
        }
        else {
            title = lua_tostring( state, 2 );
        }
    }

    MessageBox( NULL, text.c_str(), title.c_str(), MB_OK | MB_TOPMOST | MB_SETFOREGROUND );

    return 0;
}

// pause, taken from mzscript
/*
static int l_pause( lua_State* state )
{
    if ( !lua_isnumber( state, 1 ) ) {
        // this is how user defined functions produce errors
        // like built in ones: the return value from this lua_pcall
        // is nonzero, and the top of the stack is printed to the console
        lua_pushstring( state, "incorrect argument list: pause" );
        lua_error( state );
    }
    
    unsigned long lInterval = static_cast<unsigned long>( lua_tonumber( state, 1 ) );
    unsigned long lTime = GetTickCount();
    MSG msg;

    if( lInterval > 0 ) {
        while( ( GetTickCount() - lTime ) < lInterval ) {
            if( PeekMessage( &msg, NULL, 0, 0, PM_REMOVE ) ) {
                TranslateMessage( &msg );
                DispatchMessage( &msg );
            }
        }
    }

//    Sleep( static_cast< unsigned long >( lua_tonumber( state, 1 ) ) );
    
    return 0;
}
*/
// this needs to be a table with appropriate metamethods, otherwise assigning
// makes no sense when we want to support saving evars
static int l_evar( lua_State* state )
{
    if ( !lua_isstring( state, 1 ) ) {
        lua_pushstring( state, "incorrect argument list: _evar" );
        lua_error( state );
    }
    
    std::string evar = "$";
    evar += lua_tostring( state, 1 );
    evar += "$";

    char buffer[ MAX_LINE_LENGTH ] = { 0 };
    VarExpansionEx( buffer, evar.c_str(), MAX_LINE_LENGTH );

    lua_pushstring( state, buffer );
    return 1;
}

static int l_exec( lua_State* state )
{
    if ( !lua_isstring( state, 1 ) ) {
        lua_pushstring( state, "incorrect argument list: exec" );
        lua_error( state );
    }

    std::string exec_string = lua_tostring( state, 1 );
    // adding quotes if none exist, since lsexecute needs them. is this safe?
//    if ( exec_string[ 0 ] != '"' && exec_string[ exec_string.length() - 1] != '"' ) {
//        exec_string = "\"" + exec_string + "\"";
//    }

    int argc = lua_gettop( state );
    if ( argc > 1 ) {
        std::string argv;
        for ( int i = 2; i <= argc; i++ ) {
            if ( !lua_isstring( state, i ) ) {
                lua_pushstring( state, "incorrect argument list: exec" );
                lua_error( state );
            }
            argv += lua_tostring( state, i );
            argv += " ";
        }
        argv = argv.substr( 0, argv.length() - 1 );

        // do more testing here... is this a document? a folder? etc.
        LSExecuteEx( NULL, NULL, exec_string.c_str(), argv.c_str(), NULL, SW_NORMAL );
    }
    else {
        LSExecute( NULL, exec_string.c_str(), SW_NORMAL );
    }

    return 0;
}



// class member functions
Lua::Lua() 
{
    _state = lua_open();
    _load_lua_libs();
    _register_lua_functions();
    _active = true;
}

Lua::~Lua()
{
    lua_close( _state );
}

// why cant these be inline?
// lua_test.o(.text+0x217): In function `main':
// /home/doy/coding/arsphysica/lua_testing/lua_test.cpp:30: undefined reference to `Lua::set_not_active()'
// ??
bool Lua::is_active()
{
    return _active;
}

void Lua::set_active()
{
    _active = true;
    // process any scripts loaded since it was last active
    // this will produce an additional error if there was an error loading
    // a script earlier, which failed because of syntax problems, since
    // it will still be loaded, and will now be run... be sure to handle those
    // errors before setting lua active again, perhaps by loading an empty
    // function definition in first, and then calling this. if it errors here,
    // lua will *not* be set active!
    if ( _run_lua_script() ) {
        // handle error
        _lua_error( lua_tostring( _state, -1 ) );
    }
}

void Lua::set_not_active()
{
    _active = false;
}

void Lua::trigger( const char* func_name )
{
    lua_State* thread;
    if ( thread = _should_call( func_name ) ) {
        _make_lua_call( thread, func_name );
    }
}

void Lua::trigger( const char* func_name, const char* arg1 )
{
    lua_State* thread;
    if ( thread = _should_call( func_name ) ) {
        lua_checkstack( thread, 1 );
        lua_pushstring( thread, arg1 );
        _make_lua_call( thread, func_name, 1 );
    }
}

/*
void Lua::trigger( const char* func_name, lua_Number arg1 )
{
    if ( _should_call( func_name ) ) {
        lua_pushnumber( _state, arg1 );
        _make_lua_call( func_name, 1 );
    }
}

void Lua::trigger( const char* func_name, lua_Number arg1,
                          lua_Number arg2 ) 
{
    if ( _should_call( func_name ) ) {
        lua_pushnumber( _state, arg1 );
        lua_pushnumber( _state, arg2 );
        _make_lua_call( func_name, 2 );
    }
}

void Lua::trigger( const char* func_name, lua_Number arg1,
                          lua_Number arg2, lua_Number arg3 )
{
    if ( _should_call( func_name ) ) {
        lua_pushnumber( _state, arg1 );
        lua_pushnumber( _state, arg2 );
        lua_pushnumber( _state, arg3 );
        _make_lua_call( func_name, 3 );
    }
}

void Lua::trigger( const char* func_name, 
                   lua_Number arg1, lua_Number arg2, lua_Number arg3,
                   lua_Number arg4, lua_Number arg5, lua_Number arg6 )
{
    if ( _should_call( func_name ) ) {
        lua_pushnumber( _state, arg1 );
        lua_pushnumber( _state, arg2 );
        lua_pushnumber( _state, arg3 );
        lua_pushnumber( _state, arg4 );
        lua_pushnumber( _state, arg5 );
        lua_pushnumber( _state, arg6 );
        _make_lua_call( func_name, 6 );
    }
}
*/
lua_State* Lua::_should_call( const char* func_name )
{
    // first check to see if we should do any lua stuff
    if ( !_active ) {
        return NULL;
    }

    lua_checkstack( _state, 1 );
    lua_State* new_thread = lua_newthread( _state );

    // next check to see if this is a valid function name. this converts
    // the string containing the function name into the actual function,
    // if it exists, and then we test to see if it actually became a function.
    lua_checkstack( new_thread, 1 );
    lua_pushstring( new_thread, func_name );
    lua_gettable( new_thread, LUA_GLOBALSINDEX );
 
    if ( !lua_isfunction( new_thread, -1 ) ) {
        lua_pop( new_thread, 1 );
        return NULL;
    }

    // leave the function on top of the stack, since it needs to be
    // pushed before any potential arguments
    return new_thread;
}

void Lua::_make_lua_call( lua_State* thread, const char* func_name,
                          int argc /* =0 */, int retc /* =0 */ )
{
    
    // lua_pcall can return:
    // LUA_ERRRUN: runtime error
    // LUA_ERRMEM: mem allocation error
    // 
    // 4th arg here is an error handling function, should we ever choose
    // to define one

    // starting stack (_state, thread )
    // [ new_thread<thr> ] [ func ]
    // copy the thread over to the thread's stack
    // [ ] [ func, new_thread<thr> ]
    lua_checkstack( thread, 1 );
    lua_xmove( _state, thread, 1 );
    // name of global to hold the running thread
    // [ ] [ func, new_thread<thr>, "CURRENT_THREAD"<str> ]
    lua_checkstack( thread, 1 );
    lua_pushstring( thread, "CURRENT_THREAD" );
    // swap
    // [ ] [ func, "CURRENT_THREAD"<str>, new_thread<thr> ]
    lua_insert( thread, -2 );
    // create the global
    // [ ] [ func ]
    lua_rawset( thread, LUA_GLOBALSINDEX );
    
    if ( lua_resume( thread, argc ) ) {
        // put error handling/debugging code here
        // func_name is passed mainly for debugging purposes, since the
        // name has been converted to a lua function reference at this point
        // by _should_call.
        _lua_error( lua_tostring( thread, -1 ) );
    }

    // nil out the CURRENT_THREAD so it can be collected
    if ( !lua_gettop( thread ) ) {
        lua_checkstack( thread, 1 );
        lua_pushstring( thread, "CURRENT_THREAD" );
        lua_checkstack( thread, 1 );
        lua_pushnil( thread );
        lua_rawset( thread, LUA_GLOBALSINDEX );
    }

/*
    if ( lua_gettop( thread ) == 1 ) {
        if ( !lua_isstring( thread, 1 ) ) {
            _lua_error( "returned thread name was not a string" );
        }
        // time for some stack manipulation
        // initial stack:
        // [ new_thread<thr> ]
        // push name
        // [ new_thread<thr>, name<str> ]
        lua_pushstring( _state, lua_tostring( thread, 1 ) );
        // swap
        // [ name<str>, new_thread<thr> ]
        lua_insert( _state, -2 );
        // push "threads"
        // [ name<str>, new_thread<thr>, "threads"<str> ]
        lua_pushstring( _state, "threads" );
        // get global object with name "threads"
        // [ name<str>, new_thread<thr>, threads<tbl> ]
        lua_rawget( _state, -1 );
        // rot
        // [ threads<tbl>, name<str>, new_thread<thr> ]
        lua_insert( _state, -3 );
        // dup
        // [ threads<tbl>, name<str>, new_thread<thr>, new_thread<thr> ]
        lua_pushvalue( _state, -1 );
        // rot
        // [ new_thread<thr>, threads<tbl>, name<str>, new_thread<thr> ]
        lua_insert( _state, -4 );
        // threads[ name ] = new_thread
        // [ new_thread<thr>, threads<tbl> ]
        lua_rawset( _state, -3 );
        // pop (leave thread there for next call to yield())
        // [ new_thread<thr> ]
        lua_pop( _state, 1 );
    }
*/
}

void Lua::_load_lua_libs()
{
    for ( const luaL_reg* lib = _lualibs; lib->func; lib++ ) {
        lib->func( _state );
        lua_settop( _state, 0 );
    }
}

inline void Lua::_register_lua_functions()
{
    lua_register( _state, "message_box", l_message_box );
//    lua_register( _state, "pause", l_pause );
    lua_register( _state, "exec", l_exec );
    lua_register( _state, "_evar", l_evar );
}

int Lua::_run_lua_script()
{
    if ( !_active ) {
        // we dont have an error handling function, so this wont ever actually
        // happen in lua_pcall. this will actually never happen here either,
        // since the only thing that calls it right now is add_script, but
        // this is here in case that changes.
        return LUA_ERRERR;
    }
    return lua_pcall( _state, 0, LUA_MULTRET, 0 );
}

// tries to load a file. if luaL_loadfile succeds, it returns 0, and so
// _run_lua_script() will be called to refresh the symbol table.
void Lua::add_script( const char* script_name )
{
    // *** modify this to add bangs ***
    // apparently the rc parsing code leaves in quotes, so remove them
    // or lua will whine
    std::string name = script_name;
    if ( name[ 0 ] == '"' && name[ name.length() - 1 ] == '"' ) {
        name = name.substr( 1, name.length() - 2 );
    }

    // luaL_loadfile can return:
    // LUA_ERRFILE:   could not open file
    // LUA_ERRSYNTAX: syntax error
    // LUA_ERRMEM:    mem allocation error
    if ( luaL_loadfile( _state, name.c_str() ) ) {
        // other error handling
        _lua_error( lua_tostring( _state, -1 ) );
        return;
    }
    
    // if lua isnt active, we will wait until it reactivates to run this
    if ( _active ) {
        // _run_lua_script can return:
        // LUA_ERRRUN: runtime error
        // LUA_ERRMEM: mem allocation error
        if ( _run_lua_script() ) {
            // other error handling
            _lua_error( lua_tostring( _state, -1 ) );
            return;
        }
    }
    // _active is obviously true here if _run_lua_script succeeded,
    // so no need to redefine it here.
    /*
    lua_pushnil( _state );
    while ( lua_next( _state, LUA_GLOBALSINDEX ) ) {
        std::string message;
        message += lua_typename( _state, lua_type( _state, -2 ) );
        message += " - ";
        message += lua_typename( _state, lua_type( _state, -1 ) );
        message += "\n";
        if ( lua_isstring( _state, -2 ) ) {
            message += lua_tostring( _state, -2 );
        }
        else {
            message += "not a string";
        }
        message += " - ";
        if ( lua_isstring( _state, -1 ) ) {
            message += lua_tostring( _state, -1 );
        }
        else {
            message += "not a string";
        }
        MessageBox( NULL, message.c_str(), "iterating...", NULL );
        lua_pop( _state, 1 );
    }
    */
    return;
}

void Lua::_lua_error( const char* error_message )
{
    //std::cout << error_message << std::endl;
    Error( error_message );
    _active = false;
}
